@@ -60,7 +60,62 @@
   "Create a command function and store its interactive hints in
 *command-hash*. The local variable %interactivep% can be used to check
 if the command was called interactively. If it is non-NIL then it was
-called from a keybinding or from the colon command."
+called from a keybinding or from the colon command.
+
+INTERACTIVE-ARGS is a list of the following form: ((TYPE PROMPT) (TYPE PROMPT) ...)
+
+each element in INTERACTIVE-ARGS declares the type and prompt for the
+command's arguments.
+
+TYPE can be one of the following:
+
+@table @var
+@item :y-or-n
+A yes or no question returning T or NIL.
+@item :variable
+A lisp variable
+@item :function
+A lisp function
+@item :command
+A stumpwm command as a string.
+@item :key-seq
+A key sequence starting from *TOP-MAP*
+@item :window-number
+An existing window number
+@item :number
+An integer number
+@item :string
+A string
+@item :key
+A single key chord
+@item :window-name
+An existing window's name
+@item :direction
+A direction symbol. One of :UP :DOWN :LEFT :RIGHT
+@item :gravity
+A gravity symbol. One of :center :top :right :bottom :left :top-right :top-left :bottom-right :bottom-left
+@item :group
+An existing group
+@item :frame
+A frame
+@item :shell
+A shell command
+@item :rest
+The rest of the input yes to be parsed.
+@item :module
+An existing stumpwm module
+@end table
+
+Note that new argument types can be created with DEFINE-STUMPWM-TYPE.
+
+PROMPT can be string. In this case, if the corresponding argument is
+missing from an interactive call, stumpwm will use prompt for its
+value using PROMPT. If PROMPT is missing or nil, then the argument is
+considered an optional interactive argument and is not prompted for
+when missing.
+
+Alternatively, instead of specifying nil for PROMPT or leaving it
+out, an element can just be the argument type."
   (check-type name symbol)
   (let ((docstring (if (stringp (first body))
                      (first body)
@@ -107,7 +162,10 @@ alias name for the command that is only accessible interactively."
 
 (defun get-command-symbol (command)
   (if (stringp command)
-      (find-symbol (string-upcase command) :stumpwm)
+      (maphash (lambda (k v)
+                 (when (string-equal k command)
+                   (return-from get-command-symbol k)))
+               *command-hash*)
       command))
 
 (defun get-command-structure (command)
@@ -332,17 +390,18 @@ then describes the symbol."
 
 (defun select-group (screen query)
   "Attempt to match string QUERY against group number or partial name."
-  (let ((num (ignore-errors (parse-integer query)))
-        match)
-    (labels ((match (grp)
-               (let* ((name (group-name grp))
-                      (end (min (length name) (length query))))
-                 ;; try by name or number
-                 (or (string-equal name query :end1 end :end2 end)
-                     (eql (group-number grp) num)))))
-      (unless (null query)
-        (setf match (find-if #'match (screen-groups screen))))
-      match)))
+  (let ((num (ignore-errors (parse-integer query))))
+    (labels ((match-whole (grp)
+               (string-equal (group-name grp) query))
+             (match-partial (grp)
+               (let* ((end (min (length (group-name grp)) (length query))))
+                 (string-equal (group-name grp) query :end1 end :end2 end)))
+             (match-num (grp)
+               (eql (group-number grp) num)))
+      (when query
+        (or (find-if #'match-whole (screen-groups screen))
+            (find-if #'match-partial (screen-groups screen))
+            (find-if #'match-num (screen-groups screen)))))))
 
 (define-stumpwm-type :group (input prompt)
   (let ((match (select-group (current-screen)
