@@ -1,6 +1,8 @@
 ;;; MPD client & formatters for stumpwm
 ;;;
-;;; Copyright 2007 Morgan Veyret.
+;;; Copyright 2007 Morgan Veyret, Ivy Foster.
+;;;
+;;; Maintainer: Morgan Veyret
 ;;;
 ;;; This module is free software; you can redistribute it and/or modify
 ;;; it under the terms of the GNU General Public License as published by
@@ -22,19 +24,82 @@
 ;;;
 ;;; Put:
 ;;;
-;;;     (load "/path/to/mpd.lisp")
+;;;     (load-module "mpd")
 ;;;
-;;; In your ~/.stumpwmrc
+;;; ...into your ~/.stumpwmrc
 ;;;
-;;; Then you can use "%m" in your mode line format.
-;;; You can also use the various commands defined at the end of the file
+;;; Then you can use "%m" in your mode line format, as well as various commands
+;;; defined at the end of the file.
 ;;;
+;;; You can customize the modeline format (*mpd-modeline-fmt*), the status
+;;; message displayed by the command mpd-status (*mpd-status-fmt*; note that
+;;; this is also used on the modeline), and the status message displayed by the
+;;; command mpd-current-song (*mpd-current-song-fmt*). See the documentation for
+;;; *mpd-modeline-fmt* for more information.
+
 ;;; NOTES:
-;;; see http://mpd.wikia.com/wiki/Protocol_Reference for full protocol
+;;;
+;;; See http://mpd.wikia.com/wiki/Protocol_Reference for full protocol
+
+;;; TODO:
+;;;
+;;; - Implement optional shortening for formatting functions
+;;; - Set-crossfade support
+
+;;; CODE:
+
 #-(or sbcl clisp) (error "unimplemented")
 
 (in-package :stumpwm)
 
+(export '(*mpd-timeout*
+	  *mpd-collapse-album-length*
+	  *mpd-collapse-all-length*
+	  *mpd-current-song-fmt*
+	  *mpd-formatters-alist*
+	  *mpd-status-fmt*
+	  *mpd-modeline-fmt*
+	  *mpd-volume-step*
+	  *mpd-port*
+	  mpd-browse-playlist
+	  select-song-from-playlist
+	  mpd-play
+	  mpd-toggle-repeat
+	  mpd-toggle-random
+	  mpd-toggle-pause
+	  mpd-kill
+	  mpd-disconnect
+	  mpd-connect
+	  mpd-browse-artists
+	  mpd-browse-albums
+	  mpd-update
+	  mpd-clear
+	  mpd-volume-down
+	  mpd-volume-up
+	  mpd-set-volume
+	  mpd-prev
+	  mpd-next
+	  mpd-stop
+	  mpd-play-track
+	  mpd-search-genre
+	  mpd-search-album
+	  mpd-search-title
+	  mpd-search-file
+	  mpd-search-artist
+	  mpd-search-and-add-genre
+	  mpd-search-and-add-album
+	  mpd-search-and-add-title
+	  mpd-search-and-add-file
+	  mpd-search-and-add-artist
+	  mpd-add-file
+	  mpd-playlist
+	  mpd-status
+	  mpd-current-song
+	  *mpd-map*
+	  *mpd-add-map*
+	  *mpd-browse-map*
+	  *mpd-search-map*))
+
 ;;mpd client
 (defparameter *mpd-socket* nil)
 (defparameter *mpd-server*
@@ -45,6 +110,10 @@
   )
 (defparameter *mpd-port* 6600)
 
+(defvar *mpd-timeout* 50)
+
+(defvar *mpd-timer* nil)
+
 (defvar *mpd-collapse-album-length* nil)
 (defvar *mpd-collapse-all-length* nil)
 
@@ -53,7 +122,8 @@
        (handler-case (progn ,@body)
                      (error (c) (progn
                                   (message "Error with mpd connection: ~a" c)
-                                  (setf *mpd-socket* nil))))
+                                  (setf *mpd-socket* nil)
+                                  (when *mpd-timer* (cancel-timer *mpd-timer*)))))
      (message "Error: not connected to mpd~%")))
 
 (defun mpd-send (command)
@@ -70,6 +140,9 @@
   (mpd-send cmd)
   (mpd-receive))
 
+(defun mpd-format-command (fmt &rest args)
+  (mpd-send-command (apply 'format nil fmt args)))
+
 (defun mpd-termination-p (str)
   (or (mpd-error-p str)
       (mpd-ok-p str)))
@@ -102,7 +175,7 @@
 
 (defun init-mpd-connection ()
   "Connect to mpd server"
-    (setq *mpd-socket*
+    (setf *mpd-socket*
           #+clisp
         (handler-case (socket:socket-connect *mpd-port* *mpd-server*
                                              :element-type 'character)
@@ -122,19 +195,30 @@
                          (err)
                        (format t  "Error connecting to mpd: ~a~%" err))))
   (when *mpd-socket*
+    (when *mpd-timeout*
+      (setf *mpd-timer*
+            (run-with-timer *mpd-timeout* *mpd-timeout* 'mpd-ping)))
     (read-line *mpd-socket*)))
 
-(defun close-mpd-connection ()
-  "Disconnect from mpd server"
-  (with-mpd-connection
-   (close *mpd-socket*)
-   (setf *mpd-socket* nil)))
+(defun mpd-ping ()
+  (mpd-send-command "ping"))
 
+(defun mpd-search (type what &optional (exact-search nil))
+  (mpd-format-command "~a ~a \"~a\""
+                      (if exact-search "find"
+                          "search")
+                      type what))
 
+(defun mpd-add (files)
+  (loop for i in files
+     do (mpd-format-command "add \"~a\"" i)))
 
-;;mpd formatter
-(dolist (a '((#\m fmt-mpd-status)))
-  (push a *screen-mode-line-formatters*))
+;;; ------------------------------------------------------------------
+;;; Formatting
+;;; ------------------------------------------------------------------
+
+(dolist (a '((#\m mpd-modeline)))
+  (pushnew a *screen-mode-line-formatters* :test 'equal))
 
 (defparameter *mpd-current-song* nil)
 (defparameter *mpd-status* nil)
@@ -144,86 +228,198 @@
 (defun mpd-update-status ()
   (setf *mpd-status* (mpd-send-command "status")))
 
-(defun format-mpd-current-song (current-song &optional
-                                             (collapse-album nil)
-                                             (collapse-all nil))
-  (let* ((artist (assoc-value :artist current-song))
-        (album (assoc-value :album current-song))
-        (title (assoc-value :title current-song))
-         (file (assoc-value :file current-song))
+(defun mpd-get-artist ()
+  (assoc-value :artist *mpd-current-song*))
+
+(defun mpd-get-album ()
+  (assoc-value :album *mpd-current-song*))
+
+(defun mpd-get-date ()
+  (assoc-value :date *mpd-current-song*))
+
+(defun mpd-minutes-seconds (time)
+  (let ((minutes) (seconds))
+    (if (< time 60)
+	(progn
+	  (setf minutes 0)
+	  (setf seconds time))
+      (progn
+	(setf minutes (write-to-string (floor time 60)))
+	(setf seconds (rem time 60))))
+    (when (< seconds 10)
+      (setf seconds (concat "0" (write-to-string seconds))))
+    (format nil "~a:~a" minutes seconds)))
+
+(defun mpd-get-elapsed ()
+  (let* ((total (assoc-value :time *mpd-current-song*))
+	 (time (assoc-value :time *mpd-status*))
+	 (elapsed (parse-integer (subseq time 0 (- (length time) (length total) 1)))))
+    (mpd-minutes-seconds elapsed)))
+
+(defun mpd-get-length ()
+  (let ((time (parse-integer (assoc-value :time *mpd-current-song*)))
+	(minutes) (seconds))
+    (mpd-minutes-seconds time)))
+
+(defun mpd-get-status ()
+  (cond ((equal (assoc-value :state *mpd-status*) "play") "Playing")
+	((equal (assoc-value :state *mpd-status*) "pause") "Paused")
+	((equal (assoc-value :state *mpd-status*) "stop") "Stopped")))
+
+(defun mpd-get-file ()
+  (assoc-value :file *mpd-current-song*))
+
+(defun mpd-get-volume ()
+  (assoc-value :volume *mpd-status*))
+
+(defun mpd-get-xfade ()
+  (let ((xfade (assoc-value :xfade *mpd-status*)))
+    (if (> (parse-integer xfade) 0)
+	(format nil "F=~a" (assoc-value :xfade *mpd-status*)) "_")))
+
+(defun mpd-get-genre ()
+  (assoc-value :genre *mpd-current-song*))
+
+(defun mpd-get-number ()
+  (write-to-string (1+ (parse-integer (assoc-value :song *mpd-status*)))))
+
+(defun mpd-get-playlistlength ()
+  (assoc-value :playlistlength *mpd-status*))
+
+(defun mpd-repeating-p ()
+  (if (string= (assoc-value :repeat *mpd-status*) "1")
+      t nil))
+
+(defun mpd-get-repeat ()
+  (if (mpd-repeating-p) "R" "_"))
+
+(defun mpd-shuffle-p ()
+  (if (string= (assoc-value :random *mpd-status*) "1")
+      t nil))
+
+(defun mpd-get-shuffle ()
+  (if (mpd-shuffle-p) "S" "_"))
+
+(defun mpd-get-title ()
+  (assoc-value :title *mpd-current-song*))
+
+(defun mpd-get-track ()
+  (assoc-value :track *mpd-current-song*))
+
+(defun mpd-get-song-name ()
+  (let* ((artist (assoc-value :artist *mpd-current-song*))
+        (album (assoc-value :album *mpd-current-song*))
+        (title (assoc-value :title *mpd-current-song*))
+         (file (assoc-value :file *mpd-current-song*))
          (song (if (or (null artist)
-            (null album)
-            (null title))
-        (format nil "~a" file)
+                      (null album)
+                      (null title))
+                  (format nil "~a" file)
                    (format nil "~a \"~a\" - ~a"
-              artist
-              (if (and collapse-album *mpd-collapse-album-length*
-                       (> (length album) *mpd-collapse-album-length*))
-                  (concatenate 'string
-                               (subseq album 0 *mpd-collapse-album-length*)
-                               "...")
-                album)
-              title))))
-    (if (and collapse-all *mpd-collapse-all-length*
-             (> (length song) *mpd-collapse-all-length*))
+                          artist
+                          (if (and *mpd-collapse-album-length*
+                                   (> (length album) *mpd-collapse-album-length*))
+                              (concatenate 'string
+                                           (subseq album 0 *mpd-collapse-album-length*)
+                                           "...")
+                              album)
+                          title))))
+    (if (and *mpd-collapse-all-length*
+            (> (length song) *mpd-collapse-all-length*))
         (concatenate 'string (subseq song 0 *mpd-collapse-all-length*) "...")
-      song)))
+       song)))
 
-(defun format-mpd-status (status)
-  (let ((mpd-state (assoc-value :state status)))
-     (cond
-       ((equal mpd-state "play")
-        (format nil "Playing [~a;~a] (~a%)"
-                (if (equal (assoc-value :random *mpd-status*)
-                           "0")
-                    "_"
-                  "S")
-                (if (equal (assoc-value :repeat *mpd-status*)
-                           "0")
-                    "_"
-                  "R")
-                (assoc-value :volume *mpd-status*)))
-       ((equal mpd-state "pause")
-        (format nil "Paused [~a;~a]"
-                (if (equal (assoc-value :random *mpd-status*)
-                           "0")
-                    "_"
-                  "S")
-                (if (equal (assoc-value :repeat *mpd-status*)
-                           "0")
-                    "_"
-                  "R")))
-       ((equal mpd-state "stop")
-        (format nil "Stopped  [~a;~a]"
-                (if (equal (assoc-value :random *mpd-status*)
-                           "0")
-                    "_"
-                  "S")
-                (if (equal (assoc-value :repeat *mpd-status*)
-                           "0")
-                    "_"
-                  "R"))))))
-
-;;FIXME: update status based on last current song time
-;;FIXME: this means updating status on commands calls
-(defun fmt-mpd-status (ml)
+(defun mpd-modeline (ml)
   (declare (ignore ml))
   (if *mpd-socket*
       (with-mpd-connection
        (mpd-update-status)
-       (if (equal "stop" (assoc-value :state *mpd-status*))
-           (format nil "~a"
-                   (format-mpd-status *mpd-status*))
-         (progn (mpd-update-current-song)
-                (format nil "~a: ~a"
-                        (format-mpd-status *mpd-status*)
-                        (format-mpd-current-song *mpd-current-song* t t)))))
-    "Not connected"))
+       (if (equal "Stopped" (mpd-get-status))
+	   (format-expand *mpd-formatters-alist* *mpd-status-fmt*)
+	 (progn
+	   (mpd-update-current-song)
+	   (format-expand *mpd-formatters-alist* *mpd-modeline-fmt*))))))
+
+(defvar *mpd-formatters-alist*
+  '((#\a mpd-get-artist)
+    (#\A mpd-get-album)
+    (#\d mpd-get-date)
+    (#\e mpd-get-elapsed)
+    (#\f mpd-get-file)
+    (#\F mpd-get-xfade)
+    (#\g mpd-get-genre)
+    (#\l mpd-get-length)
+    (#\n mpd-get-number)
+    (#\p mpd-get-playlistlength)
+    (#\r mpd-get-repeat)
+    (#\s mpd-get-shuffle)
+    (#\S mpd-get-status)
+    (#\t mpd-get-title)
+    (#\T mpd-get-track)
+    (#\v mpd-get-volume)
+    (#\m mpd-get-song-name)))
+
+(defvar *mpd-current-song-fmt* "%a
+%A
+%t
+(%n/%p)"
+  "The default value for displaying the current song.
+For more information on valid formatters, please see the documentation
+for `*mpd-modeline-fmt*'")
+
+(defvar *mpd-status-fmt* "%S [%s;%r;%F]"
+  "The default value for displaying the current MPD status.
+For more information on valid formatters, please see the documentation
+for `*mpd-modeline-fmt*'")
+
+(defvar *mpd-modeline-fmt* "%S [%s;%r;%F]: %a - %A - %t (%n/%p)"
+  "The default value for displaying MPD information on the modeline.
+
+@table @asis
+@item %%
+A literal '%'
+@item %a
+Artist
+@item %A
+Album
+@item %d
+Date
+@item %e
+Elapsed time
+@item %f
+Filename
+@item %F
+'F=#' if crossfade is set, '_' otherwise
+@item %g
+Genre
+@item %l
+Song length
+@item %n
+Position in playlist (song Number, for a mnemonic)
+@item %p
+Total playlist length
+@item %r
+'R' if repeat is on, '_' otherwise
+@item %s
+'S' if shuffle is on, '_' otherwise
+@item %S
+'Playing' if playing, 'Paused' if paused, else 'Stopped'
+@item %t
+Title
+@item %T
+Track number (relative to the album, not the playlist)
+@item %v
+Volume
+@end table
+")
+
+;;; ------------------------------------------------------------------
+;;; Misc. commands
+;;; ------------------------------------------------------------------
 
-;;mpd commands
 (defvar *mpd-volume-step* 5)
 
-(defcommand mpd-select-song-in-playlist () ()
+(defcommand mpd-browse-playlist () ()
   (let ((status (mpd-send-command "status")))
     (labels ((pick (options)
                    (let ((selection
@@ -248,36 +444,78 @@
                (song-number (position choice result)))
           (mpd-send-command (format nil "play ~d" song-number)))))))
 
+(defcommand-alias select-song-from-playlist browse-playlist)
+
+(defcommand mpd-browse-albums () ()
+  (labels ((pick (options)
+             (let ((selection
+                    (select-from-menu (current-screen) options
+                                      "Play album" 0)))
+               (cond
+                 ((null selection)
+                  (throw 'stumpwm::error "Abort."))
+                 (t selection)))))
+    (let* ((response (mpd-send-command "list album"))
+           (result (mapcar #'cadr (remove-if (lambda (item)
+                                               (not (equal :album
+                                                           (first item))))
+                                             response))))
+      (let* ((choice (pick result)))
+        (mpd-search-and-add-album choice t)))))
+
+(defcommand mpd-browse-artists () ()
+  (labels ((pick (options)
+             (let ((selection
+                    (select-from-menu (current-screen) options
+                                      "Play artists" 0)))
+               (cond
+                 ((null selection)
+                  (throw 'stumpwm::error "Abort."))
+                 (t selection)))))
+    (let* ((response (mpd-send-command "list artist"))
+           (result (mapcar #'cadr (remove-if (lambda (item)
+                                               (not (equal :artist
+                                                           (first item))))
+                                             response))))
+      (let* ((choice (pick result)))
+        (mpd-search-and-add-artist choice t)))))
+
 (defcommand mpd-connect () ()
   (message "~a" (init-mpd-connection)))
 
 (defcommand mpd-disconnect () ()
-  (close-mpd-connection))
+  "Disconnect from mpd server"
+  (with-mpd-connection
+   (close *mpd-socket*)
+   (setf *mpd-socket* nil)
+   (when *mpd-timer* (cancel-timer *mpd-timer*))))
 
+(defcommand mpd-kill () ()
+ (mpd-send-command "kill"))
 
 (defcommand mpd-toggle-pause () ()
-  (let ((status (mpd-send-command "status")))
-    (cond
-     ((equal (assoc-value :state status)
-             "play") (mpd-send-command "pause 1"))
-     ((equal (assoc-value :state status)
-             "pause") (mpd-send-command "pause 0")))))
+  (cond
+   ((equal (mpd-get-status) "Playing")
+    (mpd-send-command "pause 1"))
+   ((equal (mpd-get-status) "Paused")
+      (mpd-send-command "pause 0"))))
 
 (defcommand mpd-toggle-random () ()
-  (let ((status (mpd-send-command "status")))
-    (cond
-     ((equal (assoc-value :random status) "0") (mpd-send-command "random 1"))
-     ((equal (assoc-value :random status) "1") (mpd-send-command "random 0")))))
+  (if (mpd-shuffle-p)
+      (mpd-send-command "random 0")
+    (mpd-send-command "random 1")))
 
 (defcommand mpd-toggle-repeat () ()
-  (let ((status (mpd-send-command "status")))
-    (cond
-     ((equal (assoc-value :repeat status) "0") (mpd-send-command "repeat 1"))
-     ((equal (assoc-value :repeat status) "1") (mpd-send-command "repeat 0")))))
+  (if (mpd-repeating-p)
+      (mpd-send-command "repeat 0")
+    (mpd-send-command "repeat 1")))
 
 (defcommand mpd-play () ()
   (mpd-send-command "play"))
 
+(defcommand mpd-play-track (track) ((:number "Track: "))
+  (mpd-send-command (concat "play " (write-to-string (1- track)))))
+
 (defcommand mpd-stop () ()
   (mpd-send-command "stop"))
 
@@ -303,11 +541,22 @@
 (defcommand mpd-clear () ()
   (mpd-send-command "clear"))
 
-(defcommand mpd-update () ()
-  (message "~a" (mpd-send-command "update")))
+(defcommand mpd-update (&optional (path nil)) ()
+  (if path
+      (message "~a" (mpd-format-command "update ~a" path))
+      (message "~a" (mpd-send-command "update"))))
 
 (defcommand mpd-current-song () ()
-  (message "~a" (format-mpd-current-song (mpd-send-command "currentsong"))))
+  (mpd-update-current-song)
+  (mpd-update-status)
+  (message "~a"
+	   (format-expand *mpd-formatters-alist* *mpd-current-song-fmt*)))
+
+(defcommand mpd-status () ()
+  (mpd-update-status)
+  (mpd-update-current-song)
+  (message "~a"
+	   (format-expand *mpd-formatters-alist* *mpd-status-fmt*)))
 
 (defcommand mpd-playlist () ()
   (let* ((response (mpd-send-command "playlistinfo"))
@@ -325,83 +574,76 @@
   (mpd-send-command (format nil "add \"~a\"" file)))
 
 ;;search and add commands
-(defcommand mpd-search-and-add-artist (what)
+(defcommand mpd-search-and-add-artist (what &optional (exact-search nil))
   ((:rest "Search & add artist to playlist: "))
-  (let* ((response (mpd-send-command (format nil "search artist \"~a\"" what)))
+  (let* ((response (mpd-search "artist" what exact-search))
          (result (mapcar #'cadr (remove-if (lambda (item)
                                              (not (equal :file
                                                          (first item))))
                                            response))))
-    (loop for i in result
-          do (mpd-send-command (format nil "add \"~a\"" i)))
+    (mpd-add result)
     (if (< (length result) 80)
         (message "Added ~a files: ~%^7*~{~a~%~}"
                  (length result)
                  result)
       (message "~a files added" (length result)))))
 
-(defcommand mpd-search-and-add-file (what)
+(defcommand mpd-search-and-add-file (what &optional (exact-search nil))
   ((:rest "Search & add file to playlist: "))
-  (let* ((response (mpd-send-command (format nil "search file \"~a\"" what)))
+  (let* ((response (mpd-search "file" what exact-search))
          (result (mapcar #'cadr (remove-if (lambda (item)
                                              (not (equal :file
                                                          (first item))))
                                            response))))
-    (loop for i in result
-          do (mpd-send-command (format nil "add \"~a\"" i)))
+    (mpd-add result)
     (if (< (length result) 80)
         (message "Added ~a files: ~%^7*~{~a~%~}"
                  (length result)
                  result)
       (message "~a files added" (length result)))))
 
-(defcommand mpd-search-and-add-title (what)
+(defcommand mpd-search-and-add-title (what &optional (exact-search nil))
   ((:rest "Search & add title to playlist: "))
-  (let* ((response (mpd-send-command (format nil "search title \"~a\"" what)))
+  (let* ((response (mpd-search "title" what exact-search))
          (result (mapcar #'cadr (remove-if (lambda (item)
                                              (not (equal :file
                                                          (first item))))
                                            response))))
-    (loop for i in result
-          do (mpd-send-command (format nil "add \"~a\"" i)))
+    (mpd-add result)
     (if (< (length result) 80)
         (message "Added ~a files: ~%^7*~{~a~%~}"
                  (length result)
                  result)
       (message "~a files added" (length result)))))
 
-(defcommand mpd-search-and-add-album (what)
+(defcommand mpd-search-and-add-album (what &optional (exact-search nil))
   ((:rest "Search & add album to playlist: "))
-  (let* ((response (mpd-send-command (format nil "search album \"~a\"" what)))
+  (let* ((response (mpd-search "album" what exact-search))
          (result (mapcar #'cadr (remove-if (lambda (item)
                                              (not (equal :file
                                                          (first item))))
                                            response))))
-    (loop for i in result
-          do (mpd-send-command (format nil "add \"~a\"" i)))
+    (mpd-add result)
     (if (< (length result) 80)
         (message "Added ~a files: ~%^7*~{~a~%~}"
                  (length result)
                  result)
       (message "~a files added" (length result)))))
 
-(defcommand mpd-search-and-add-genre (what)
+(defcommand mpd-search-and-add-genre (what &optional (exact-search nil))
   ((:rest "Search & add genre to playlist: "))
-  (let* ((response (mpd-send-command (format nil "search genre \"~a\"" what)))
+  (let* ((response (mpd-search "genre" what exact-search))
          (result (mapcar #'cadr (remove-if (lambda (item)
                                              (not (equal :file
                                                          (first item))))
                                            response))))
-    (loop for i in result
-          do (mpd-send-command (format nil "add \"~a\"" i)))
+    (mpd-add result)
     (if (< (length result) 80)
         (message "Added ~a files: ~%^7*~{~a~%~}"
                  (length result)
                  result)
       (message "~a files added" (length result)))))
 
-
-
 ;;search commands
 (defcommand mpd-search-artist (what) ((:rest "Search artist: "))
   (mpd-send-command (format nil "search artist \"~a\"" what)))
@@ -425,6 +667,13 @@
         (define-key m (kbd "g") "mpd-search-genre")
         m))
 
+(setf *mpd-browse-map*
+      (let ((m (make-sparse-keymap)))
+        (define-key m (kbd "p") "mpd-browse-playlist")
+        (define-key m (kbd "l") "mpd-browse-albums")
+        (define-key m (kbd "a") "mpd-browse-artists")
+        m))
+
 (setf *mpd-add-map*
       (let ((m (make-sparse-keymap)))
         (define-key m (kbd "a") "mpd-search-and-add-artist")
@@ -440,18 +689,22 @@
         (define-key m (kbd "SPC") "mpd-toggle-pause")
         (define-key m (kbd "s") "mpd-toggle-random")
         (define-key m (kbd "r") "mpd-toggle-repeat")
+        (define-key m (kbd "S") "mpd-current-song")
         (define-key m (kbd "p") "mpd-play")
+        (define-key m (kbd "q") "mpd-browse-playlist")
         (define-key m (kbd "o") "mpd-stop")
         (define-key m (kbd "m") "mpd-next")
         (define-key m (kbd "l") "mpd-prev")
         (define-key m (kbd "c") "mpd-clear")
+        (define-key m (kbd "x") "mpd-connect")
+        (define-key m (kbd "k") "mpd-kill")
         (define-key m (kbd "u") "mpd-update")
         (define-key m (kbd "a") "mpd-search-and-add-artist")
-        (define-key m (kbd "q") "mpd-select-song-in-playlist")
         (define-key m (kbd "z") "mpd-playlist")
         (define-key m (kbd "v") "mpd-set-volume")
         (define-key m (kbd "e") "mpd-volume-up")
         (define-key m (kbd "d") "mpd-volume-down")
         (define-key m (kbd "S") '*mpd-search-map*)
+        (define-key m (kbd "b") '*mpd-browse-map*)
         (define-key m (kbd "A") '*mpd-add-map*)
-        m))
\ No newline at end of file
+        m))
