@@ -263,9 +263,10 @@ The Caller is responsible for setting up the input focus."
                     (handle-keymap cmd code state key-seq nil update-fn)
                  (when grab (ungrab-pointer)))))
             (t (values cmd key-seq)))
-          (if (find key (list (kbd "?")
-                              (kbd "C-h"))
-                    :test 'equalp)
+          (if (and (find key (list (kbd "?")
+                                   (kbd "C-h"))
+                         :test 'equalp)
+                   keymap)
               (progn (display-keybinding keymap) (values t key-seq))
               (values nil key-seq))))))
 
@@ -281,7 +282,7 @@ The Caller is responsible for setting up the input focus."
                ((eq cmd t))
                (cmd
                 (unmap-message-window (current-screen))
-                (interactive-command cmd) t)
+                (eval-command cmd t) t)
                (t (message "~{~a ~}not bound." (mapcar 'print-key (nreverse key-seq))))))))))
 
 (defun bytes-to-window (bytes)
@@ -304,7 +305,7 @@ chunks."
                         (interactive-p (car data))
                         (cmd (map 'string 'code-char (nbutlast (cdr data)))))
                    (declare (ignore interactive-p))
-                   (interactive-command cmd)
+                   (eval-command cmd)
                    (xlib:change-property win :rp_command_result (map 'list 'char-code "0TODO") :string 8)
                    (xlib:display-finish-output *display*)))
                bytes-after)))
@@ -321,7 +322,7 @@ chunks."
           (*executing-stumpwm-command* t))
       (setf (screen-last-msg screen) '()
             (screen-last-msg-highlights screen) '())
-      (interactive-command cmd)
+      (eval-command cmd)
       (xlib:change-property win :stumpwm_command_result
                             (string-to-bytes (format nil "~{~{~a~%~}~}" (nreverse (screen-last-msg screen))))
                             :string 8)
@@ -335,12 +336,19 @@ chunks."
     (when (register-urgent-window window)
       (run-hook-with-args *urgent-window-hook* window))))
 
+(defun safe-atom-name (n)
+  "Return the name of the atom with atom-id N or nil if there isn't one."
+  (handler-case
+      (xlib:atom-name *display* n)
+    (xlib:atom-error ()
+      nil)))
+
 (defun safe-bytes-to-atoms (list)
   "Return a list of atoms from list. Any number that cannot be
 converted to an atom is removed."
   (loop for p in list
         when (typep p '(unsigned-byte 29))
-        collect (xlib:atom-name *display* p)))
+        collect (safe-atom-name p)))
 
 (defun update-window-properties (window atom)
   (case atom
@@ -349,7 +357,7 @@ converted to an atom is removed."
      ;; Let the mode line know about the new name.
      (update-all-mode-lines))
     (:wm_normal_hints
-     (setf (window-normal-hints window) (xlib:wm-normal-hints (window-xwin window))
+     (setf (window-normal-hints window) (get-normalized-normal-hints (window-xwin window))
            (window-type window) (xwin-type (window-xwin window)))
      (dformat 4 "new hints: ~s~%" (window-normal-hints window))
      (maximize-window window))
@@ -539,7 +547,7 @@ converted to an atom is removed."
              ;; Sometimes the number cannot be converted to an atom, so skip them.
              (unless (or (= p 0)
                          (not (typep p '(unsigned-byte 29))))
-               (case (xlib:atom-name *display* p)
+               (case (safe-atom-name p)
                  (:_NET_WM_STATE_DEMANDS_ATTENTION
                   (case action
                     (1
@@ -547,7 +555,7 @@ converted to an atom is removed."
                     (2
                      (unless (find-wm-state window :_NET_WM_STATE_DEMANDS_ATTENTION)
                        (add-wm-state window :_NET_WM_STATE_DEMANDS_ATTENTION))))
-                  (maybe-set-urgency window))
+                  (maybe-set-urgency our-window))
                (:_NET_WM_STATE_FULLSCREEN
                 (update-fullscreen our-window action)))))))))
   (:_NET_MOVERESIZE_WINDOW
@@ -580,7 +588,8 @@ the window in it's frame."
   (when (and window (eq mode :normal) (eq *mouse-focus-policy* :sloppy))
     (let ((win (find-window window)))
       (when (and win (find win (top-windows)))
-        (focus-all win)))))
+        (focus-all win)
+        (update-all-mode-lines)))))
 
 (define-stump-event-handler :button-press (window code x y child time)
   ;; Pass click to client
@@ -593,13 +602,15 @@ the window in it's frame."
          (let* ((group (screen-current-group screen))
                 (frame (find-frame group x y)))
            (when frame
-             (focus-frame group frame))))
+             (focus-frame group frame)
+             (update-all-mode-lines))))
        (run-hook-with-args *root-click-hook* screen code x y))
       ((setf ml (find-mode-line-window window))
        (run-hook-with-args *mode-line-click-hook* ml code x y))
       ((setf win (find-window-by-parent window (visible-windows)))
        (when (eq *mouse-focus-policy* :click)
-         (focus-all win))))))
+         (focus-all win)
+         (update-all-mode-lines))))))
 
 ;; Handling event :KEY-PRESS
 ;; (:DISPLAY #<XLIB:DISPLAY :0 (The X.Org Foundation R60700000)> :EVENT-KEY :KEY-PRESS :EVENT-CODE 2 :SEND-EVENT-P NIL :CODE 45 :SEQUENCE 1419 :TIME 98761213 :ROOT #<XLIB:WINDOW :0 96> :WINDOW #<XLIB:WINDOW :0 6291484> :EVENT-WINDOW #<XLIB:WINDOW :0 6291484> :CHILD
@@ -609,8 +620,20 @@ the window in it's frame."
 (defun handle-event (&rest event-slots &key display event-key &allow-other-keys)
   (declare (ignore display))
   (dformat 1 ">>> ~S~%" event-key)
-  (let ((eventfn (gethash event-key *event-fn-table*)))
+  (let ((eventfn (gethash event-key *event-fn-table*))
+        (win (getf event-slots :window)))
     (when eventfn
+      ;; XXX: In both the clisp and sbcl clx libraries, sometimes what
+      ;; should be a window will be a pixmap instead. In this case, we
+      ;; need to manually translate it to a window to avoid breakage
+      ;; in stumpwm. So far the only slot that seems to be affected is
+      ;; the :window slot for configure-request and reparent-notify
+      ;; events. It appears as though the hash table of XIDs and clx
+      ;; structures gets out of sync with X or perhaps X assigns a
+      ;; duplicate ID for a pixmap and a window.
+      (when (and win (not (xlib:window-p win)))
+        (dformat 10 "Pixmap Workaround! ~s should be a window!~%" win)
+        (setf (getf event-slots :window) (make-xlib-window win)))
       (handler-case
           (progn
             ;; This is not the stumpwm top level, but if the restart
