@@ -113,7 +113,10 @@ T (default) then also focus the frame."
       (if w
           (raise-window w)
           (mapc 'hide-window (frame-windows g f))))
-    (when focus
+    ;; If raising a window in the current frame we must focus it or
+    ;; the group and screen will get out of sync.
+    (when (or focus
+              (eq (tile-group-current-frame g) f))
       (focus-frame g f))
     (when (and w (not (window-modal-p w)))
       (raise-modals-of w))))
@@ -158,8 +161,7 @@ T (default) then also focus the frame."
   (tree-accum-fn (tile-group-frame-head group head) 'nconc 'list))
 
 (defun find-free-frame-number (group)
-  (find-free-number (mapcar (lambda (f) (frame-number f))
-                            (group-frames group))))
+  (find-free-number (mapcar 'frame-number (group-frames group))))
 
 (defun choose-new-frame-window (frame group)
   "Find out what window should go in a newly created frame."
@@ -170,9 +172,9 @@ T (default) then also focus the frame."
     (when win
       (setf (window-frame win) frame))))
 
-(defun split-frame-h (group p)
+(defun split-frame-h (group p ratio)
   "Return 2 new frames. The first one stealing P's number and window"
-  (let* ((w (truncate (/ (frame-width p) 2)))
+  (let* ((w (truncate (* (frame-width p) ratio)))
          (h (frame-height p))
          (f1 (make-frame :number (frame-number p)
                          :x (frame-x p)
@@ -190,10 +192,10 @@ T (default) then also focus the frame."
     (run-hook-with-args *new-frame-hook* f2)
     (values f1 f2)))
 
-(defun split-frame-v (group p)
+(defun split-frame-v (group p ratio)
   "Return 2 new frames. The first one stealing P's number and window"
   (let* ((w (frame-width p))
-         (h (truncate (/ (frame-height p) 2)))
+         (h (truncate (* (frame-height p) ratio)))
          (f1 (make-frame :number (frame-number p)
                          :x (frame-x p)
                          :y (frame-y p)
@@ -485,13 +487,18 @@ either :width or :height"
   (check-type amount integer)
   ;; (check-type dim (member :width :height))
   (labels ((max-amount (parent node min dim-fn)
-             (dformat 10 "max ~@{~a~^ ~}~%" parent node min dim-fn)
-             (if parent
-                 (- (funcall dim-fn parent)
-                    (funcall dim-fn node)
-                    (* min (1- (length parent))))
-                 ;; no parent means the frame can't get any bigger.
-                 0)))
+             (let ((right-sibling (cadr (member node parent)))
+                   (left-sibling (cadr (member node (reverse parent)))))
+
+               (dformat 10 "max ~@{~a~^ ~}~%" parent node min dim-fn right-sibling left-sibling)
+               (if parent
+                   (cond (right-sibling
+                          (max 0 (- (funcall dim-fn right-sibling) min)))
+                         (left-sibling
+                          (max 0 (- (funcall dim-fn left-sibling) min)))
+                         (t 0))
+                   ;; no parent means the frame can't get any bigger.
+                   0))))
     (let* ((tree (tile-group-frame-tree group))
            (parent (tree-parent tree frame))
            (gparent (tree-parent tree parent))
@@ -578,7 +585,7 @@ depending on the tree's split direction."
      (tree-iterate i (lambda (leaf)
                        (sync-frame-windows group leaf))))))
 
-(defun split-frame (group how)
+(defun split-frame (group how &optional (ratio 1/2))
   "split the current frame into 2 frames. return T if it succeeded. NIL otherwise."
   (check-type how (member :row :column))
   (let* ((frame (tile-group-current-frame group))
@@ -591,7 +598,7 @@ depending on the tree's split direction."
       (multiple-value-bind (f1 f2) (funcall (if (eq how :column)
                                                 'split-frame-h
                                                 'split-frame-v)
-                                            group frame)
+                                            group frame ratio)
         (setf (tile-group-frame-head group head)
               (if (atom (tile-group-frame-head group head))
                   (list f1 f2)
@@ -701,12 +708,12 @@ windows used to draw the numbers in. The caller must destroy them."
 "Split the current frame into 2 frames, one on top of the other."
   (split-frame-in-dir (current-group) :row))
 
-(defcommand remove-split (&optional (group (current-group))) ()
-"Remove the current frame in the specified group (defaults to current
-group). Windows in the frame are migrated to the frame taking up its
+(defcommand remove-split (&optional (group (current-group)) (frame (tile-group-current-frame group))) ()
+"Remove the specified frame in the specified group (defaults to current
+group, current frame). Windows in the frame are migrated to the frame taking up its
 space."
-  (let* ((frame (tile-group-current-frame group))
-         (head (frame-head group frame))
+  (let* ((head (frame-head group frame))
+         (current (tile-group-current-frame group))
          (tree (tile-group-frame-head group head))
          (s (closest-sibling (list tree) frame))
          ;; grab a leaf of the siblings. The siblings doesn't have to be
@@ -731,14 +738,16 @@ space."
           ;; Unsplit
           (setf (tile-group-frame-head group head) (remove-frame tree frame))
           ;; update the current frame and sync all windows
-          (setf (tile-group-current-frame group) l)
+          (when (eq frame current)
+            (setf (tile-group-current-frame group) l))
           (tree-iterate tree
                         (lambda (leaf)
                           (sync-frame-windows group leaf)))
-          (frame-raise-window group l (frame-window l))
+          (frame-raise-window group l (frame-window l) nil)
           (when (frame-window l)
             (update-window-border (frame-window l)))
-          (show-frame-indicator group)))))
+          (when (eq frame current)
+            (show-frame-indicator group))))))
 
 (defcommand-alias remove remove-split)
 
@@ -746,7 +755,7 @@ space."
   "Delete all the frames but the current one and grow it to take up the entire head."
   (let* ((screen (current-screen))
          (group (screen-current-group screen))
-         (win (frame-window (tile-group-current-frame group)))
+         (win (group-current-window group))
          (head (current-head group))
          (frame (copy-frame head)))
     (if (atom (tile-group-frame-head group head))
@@ -906,7 +915,7 @@ jump to that frame."
             ;; Two edges are neighbours if they have the same offset and their starts and ends
             ;; overlap.  We want to find the neighbour that overlaps the most.
             (when (and (= src-offset offset)
-                       (>= overlap best-overlap))
+                       (> overlap best-overlap))
               (setf best-frame f)
               (setf best-overlap overlap))))))
     best-frame))
