@@ -93,14 +93,19 @@ modifier. Most of the time numlock just gets in the way."
     (when with-numlock (setf mods (append (modifiers-numlock *modifiers*) mods)))
     (apply 'xlib:make-state-mask mods)))
 
-(define-condition kbd-parse ()
-  () (:documentation "Raised when a kbd string failed to parse."))
+(defun report-kbd-parse-error (c stream)
+  (format stream "Failed to parse key string: ~s" (slot-value c 'string)))
+
+(define-condition kbd-parse-error (stumpwm-error)
+  ((string :initarg :string))
+  (:report report-kbd-parse-error)
+  (:documentation "Raised when a kbd string failed to parse."))
 
 (defun parse-mods (mods end)
   "MODS is a sequence of <MOD CHAR> #\- pairs. Return a list suitable
 for passing as the last argument to (apply #'make-key ...)"
   (unless (evenp end)
-    (signal 'kbd-parse))
+    (signal 'kbd-parse-error :string mods))
   (apply #'nconc (loop for i from 0 below end by 2
                        if (char/= (char mods (1+ i)) #\-)
                        do (signal 'kbd-parse)
@@ -111,18 +116,18 @@ for passing as the last argument to (apply #'make-key ...)"
                                  (#\H (list :hyper t))
                                  (#\s (list :super t))
                                  (#\S (list :shift t))
-                                 (t (signal 'kbd-parse))))))
+                                 (t (signal 'kbd-parse-error :string mods))))))
 
 (defun parse-key (string)
-  "Parse STRING and return a key structure."
-  ;; FIXME: we want to return NIL when we get a kbd-parse error
-  ;;(ignore-errors
+  "Parse STRING and return a key structure. Raise an error of type
+kbd-parse if the key failed to parse."
   (let* ((p (when (> (length string) 2)
               (position #\- string :from-end t :end (- (length string) 1))))
          (mods (parse-mods string (if p (1+ p) 0)))
          (keysym (stumpwm-name->keysym (subseq string (if p (1+ p) 0)))))
-    (and keysym
-         (apply 'make-key :keysym keysym mods))))
+    (if keysym
+        (apply 'make-key :keysym keysym mods)
+        (signal 'kbd-parse-error :string string))))
 
 (defun parse-key-seq (keys)
   "KEYS is a key sequence. Parse it and return the list of keys."
@@ -153,22 +158,26 @@ others."
   (format nil "^5*~{~a~^ ~}^n" (mapcar 'print-key seq)))
 
 (defun define-key (map key command)
-  "Add a keybinding mapping the key, @var{key}, to the command,
-@var{command}, in the specified keymap. For example,
+  "Add a keybinding mapping for the key, @var{key}, to the command,
+@var{command}, in the specified keymap. If @var{command} is nil, remove an
+exising binding.  For example,
 
 @example
 \(stumpwm:define-key stumpwm:*root-map* (stumpwm:kbd \"C-z\") \"echo Zzzzz...\")
 @end example
 
 Now when you type C-t C-z, you'll see the text ``Zzzzz...'' pop up."
-  (setf (gethash key map) command)
+  (if command
+      (setf (gethash key map) command)
+      (remhash key map))
   ;; We need to tell the X server when changing the top-map bindings.
   (when (eq map *top-map*)
     (sync-keys)))
 
+;; Not really needed.  Keep it for backward compatibility.
 (defun undefine-key (map key)
   "Clear the key binding in the specified keybinding."
-  (remhash key map))
+  (define-key map key nil))
 
 (defun lookup-key-sequence (kmap key-seq)
   "Return the command bound to the key sequenc, KEY-SEQ, in keymap KMAP."
