@@ -1,4 +1,4 @@
-;; Copyright (C) 2006 Shawn Betts
+;; Copyright (C) 2006-2008 Shawn Betts
 ;;
 ;;  This file is part of stumpwm.
 ;;
@@ -28,8 +28,11 @@
 	  *mode-line-position*
 	  *mode-line-timeout*
 	  *screen-mode-line-format*
+	  *screen-mode-line-formatters*
+          add-screen-mode-line-formatter
 	  enable-mode-line
-	  toggle-mode-line))
+	  toggle-mode-line
+	  bar-zone-color))
 
 (defstruct mode-line
   screen
@@ -68,8 +71,8 @@
   "The mode line border color.")
 
 (defvar *screen-mode-line-format* "[^B%n^b] %W"
-  "This variable describes what will be displayed on the modeline for each screen. 
-Turn it on with the function TOGGLE-MODE-LINE or the mode-line command. 
+  "This variable describes what will be displayed on the modeline for each screen.
+Turn it on with the function TOGGLE-MODE-LINE or the mode-line command.
 
 It is a list where each element may be a string, a symbol, or a list.
 
@@ -100,7 +103,8 @@ List the groups using @var{*group-format*}
                                         (#\g fmt-group-list)
                                         (#\h fmt-head)
                                         (#\n fmt-group)
-                                        (#\W fmt-head-window-list))
+                                        (#\W fmt-head-window-list)
+                                        (#\u fmt-urgent-window-list))
   "An alist containing format character format function pairs for
 formatting screen mode-lines. functions are passed the screen's
 current group.")
@@ -123,10 +127,26 @@ timer.")
 
 ;;; Formatters
 
+(defun add-screen-mode-line-formatter (character fmt-fun)
+  "Add a format function to a format character (or overwrite an existing one)."
+  (setf *screen-mode-line-formatters*
+        (cons (list character fmt-fun)
+              (remove character *screen-mode-line-formatters* :key #'first))))
+
 ;; All mode-line formatters take the mode-line they are being invoked from
 ;; as the first argument. Additional arguments (everything between the first
 ;; ',' and the ';' are provided as strings [not yet implemented]).
 
+(defun fmt-urgent-window-list (ml)
+  "Using *window-format*, return a 1 line list of the urgent windows, space seperated."
+   (format nil "~{~a~^ ~}"
+          (mapcar (lambda (w)
+                    (let ((str (format-expand *window-formatters* *window-format* w)))
+                      (if (eq w (current-window))
+                          (fmt-highlight str)
+                          str)))
+                  (screen-urgent-windows (mode-line-screen ml)))))
+
 (defun fmt-window-list (ml)
    "Using *window-format*, return a 1 line list of the windows, space seperated."
   (format nil "~{~a~^ ~}"
@@ -167,16 +187,16 @@ timer.")
 (defvar *bar-hi-color* "^B^3*")
 (defvar *bar-crit-color* "^B^1*")
 
-(defun bar-zone-color (percent)
-  (concatenate 'string
-               (cond
-                 ((>= percent 90)
-                  *bar-crit-color*)
-                 ((>= percent 50)
-                  *bar-hi-color*)
-                 ((>= percent 20)
-                  *bar-med-color*)
-                 (t ""))))
+(defun bar-zone-color (amount &optional (med 20) (hi 50) (crit 90) reverse)
+  "Return a color command based on the magnitude of the argument. If
+the limits for the levels aren't specified, they default to sensible
+values for a percentage. With reverse, lower numbers are more
+critical."
+  (labels ((past (n) (funcall (if reverse #'<= #'>=) amount n)))
+    (cond ((past crit) *bar-crit-color*)
+          ((past hi) *bar-hi-color*)
+          ((past med) *bar-med-color*)
+          (t ""))))
 
 (defun repeat (n char)
  (make-string n :initial-element char))
@@ -243,7 +263,7 @@ timer.")
 (defun resize-mode-line (ml)
   (when (eq (mode-line-mode ml) :stump)
     ;; This is a StumpWM mode-line
-    (setf (xlib:drawable-height (mode-line-window ml)) (+ (font-height (xlib:gcontext-font (mode-line-gc ml))) (* *mode-line-pad-y* 2))))
+    (setf (xlib:drawable-height (mode-line-window ml)) (+ (* (1+ (count #\Newline (mode-line-contents ml) :test #'equal)) (font-height (xlib:gcontext-font (mode-line-gc ml)))) (* *mode-line-pad-y* 2))))
   (setf (xlib:drawable-width (mode-line-window ml)) (- (frame-width (mode-line-head ml)) (* 2 (xlib:drawable-border-width (mode-line-window ml))))
         (xlib:drawable-height (mode-line-window ml)) (min (xlib:drawable-height (mode-line-window ml)) (/ (head-height (mode-line-head ml)) 4))
         (mode-line-height ml) (+ (xlib:drawable-height (mode-line-window ml)) (* 2 (xlib:drawable-border-width (mode-line-window ml))))
@@ -256,13 +276,17 @@ timer.")
 (defun frame-display-y (group frame)
   "Return a Y for frame that doesn't overlap the mode-line."
   (let* ((head (frame-head group frame))
-         (ml (head-mode-line head)))
+         (ml (head-mode-line head))
+	 (head-y (frame-y head))
+	 (rel-frame-y (- (frame-y frame) head-y)))
     (if (and ml (not (eq (mode-line-mode ml) :hidden)))
         (case (mode-line-position ml)
           (:top
-           (+ (mode-line-height ml) (round (* (frame-y frame) (mode-line-factor ml)))))
+           (+ head-y
+	      (+ (mode-line-height ml) (round (* rel-frame-y (mode-line-factor ml))))))
           (:bottom
-           (round (* (frame-y frame) (mode-line-factor ml)))))
+           (+ head-y
+	      (round (* rel-frame-y (mode-line-factor ml))))))
         (frame-y frame))))
 
 (defun frame-display-height (group frame)
@@ -345,11 +369,11 @@ timer.")
            (string (mode-line-format-string ml)))
       (when (or force (not (string= (mode-line-contents ml) string)))
         (setf (mode-line-contents ml) string)
-        (xlib:display-finish-output *display*)
-        (let ((width (render-strings (mode-line-screen ml) (mode-line-cc ml)
-                                     *mode-line-pad-x*     *mode-line-pad-y* (list string) '())))
-          ;; Just clear what we need to. This reduces flicker.
-          (xlib:clear-area (mode-line-window ml) :x (+ *mode-line-pad-x* width)))))))
+        (resize-mode-line ml)
+        (xlib:clear-area (mode-line-window ml))
+        (render-strings (mode-line-screen ml) (mode-line-cc ml)
+                        *mode-line-pad-x*     *mode-line-pad-y*
+                        (split-string string (string #\Newline)) '())))))
 
 (defun find-mode-line-window (xwin)
   (dolist (s *screen-list*)
@@ -493,6 +517,6 @@ timer.")
       (when (head-mode-line head)
         (toggle-mode-line screen head))))
 
-(define-stumpwm-command "mode-line" ()
+(defcommand mode-line () ()
   "A command to toggle the mode line visibility."
   (toggle-mode-line (current-screen) (current-head)))
