@@ -107,14 +107,17 @@
           *data-dir*
           add-hook
           clear-window-placement-rules
+          data-dir-file
           dformat
           define-frame-preference
+          redirect-all-output
           remove-hook
           run-hook
           run-hook-with-args
           split-string
 	  with-restarts-menu
-          with-data-file))
+          with-data-file
+	  move-to-head))
 
 
 ;;; Message Timer
@@ -537,11 +540,12 @@ restart from a menu of possible restarts. If a restart is not
 chosen, resignal the error."
   (let ((c (gensym)))
     `(handler-bind
-      ((error
-        (lambda (,c)
-          (restarts-menu ,c)
-          (signal ,c))))
-      ,@body)))
+         ((warning #'muffle-warning)
+          ((or serious-condition error)
+           (lambda (,c)
+             (restarts-menu ,c)
+             (signal ,c))))
+       ,@body)))
 
 ;;; Hook functionality
 
@@ -553,7 +557,7 @@ chosen, resignal the error."
             (dolist (fn hook)
               (with-simple-restart (continue-hooks "Continue running the remaining hooks.")
                 (apply fn args)))))
-    (error (c) (message "^B^1*Error on hook ^b~S^B!~% ^n~A" hook c) (values nil c))))
+    (t (c) (message "^B^1*Error on hook ^b~S^B!~% ^n~A" hook c) (values nil c))))
 
 (defun run-hook (hook)
   "Call each function in HOOK."
@@ -683,7 +687,7 @@ output directly to a file.")
 (defun dformat (level fmt &rest args)
   (when (>= *debug-level* level)
     (multiple-value-bind (sec m h) (decode-universal-time (get-universal-time))
-      (format *debug-stream* "~d:~d:~d " h m sec))
+      (format *debug-stream* "~2,'0d:~2,'0d:~2,'0d " h m sec))
     ;; strip out non base-char chars quick-n-dirty like
     (write-string (map 'string (lambda (ch)
                                  (if (typep ch 'base-char)
@@ -691,6 +695,28 @@ output directly to a file.")
                        (apply 'format nil fmt args))
                   *debug-stream*)))
 
+(defvar *redirect-stream* nil
+  "This variable Keeps track of the stream all output is sent to when
+`redirect-all-output' is called so if it changes we can close it
+before reopening.")
+
+(defun redirect-all-output (file)
+  "Elect to redirect all output to the specified file. For instance,
+if you want everything to go to ~/stumpwm.d/debug-output.txt you would
+do:
+
+@example
+(redirect-all-output (data-dir-file \"debug-output\" \"txt\"))
+@end example
+"
+  (when (typep *redirect-stream* 'file-stream)
+    (close *redirect-stream*))
+  (setf *redirect-stream* (open file :direction :output :if-exists :append :if-does-not-exist :create)
+        *error-output*    *redirect-stream*
+        *standard-output* *redirect-stream*
+        *trace-output*    *redirect-stream*
+        *debug-stream*    *redirect-stream*))
+
 ;;; 
 ;;; formatting routines
 
@@ -734,7 +760,10 @@ output directly to a file.")
                               (#\c window-class)
                               (#\i window-res)
                               (#\r window-role)
-                              (#\m fmt-window-marked))
+                              (#\m fmt-window-marked)
+                              (#\h window-height)
+                              (#\w window-width)
+                              (#\g gravity-for-window))
   "an alist containing format character format function pairs for formatting window lists.")
 
 (defvar *window-format* "%m%n%s%50t"
@@ -761,6 +790,9 @@ Note, a prefix number can be used to crop the argument to a specified
 size. For instance, @samp{%20t} crops the window's title to 20
 characters.")
 
+(defvar *window-info-format* "%wx%h %n (%t)"
+  "The format used in the info command. @xref{*window-format*} for formatting details.")
+
 (defvar *group-formatters* '((#\n group-number)
                              (#\s fmt-group-status)
                              (#\t group-name))
@@ -966,6 +998,7 @@ The window's title must match @var{title}.
     `(dolist (,x ',frame-rules)
        ;; verify the correct structure
        (destructuring-bind (frame-number raise lock &rest keys &key class instance type role title) ,x
+         (declare (ignore class instance type role title))
          (push (list* ,target-group frame-number raise lock keys) *window-placement-rules*)))))
 
 (defun clear-window-placement-rules ()
@@ -998,13 +1031,11 @@ input focus is transfered to the window you click on.")
 
 (defmacro with-focus (xwin &body body)
   "Set the focus to xwin, do body, then restore focus"
-  (let ((focus (gensym "FOCUS"))
-        (revert (gensym "REVERT")))
-    `(multiple-value-bind (,focus ,revert) (xlib:input-focus *display*)
-       (xlib:set-input-focus *display* ,xwin :pointer-root)
-       (unwind-protect
-            (progn ,@body)
-         (xlib:set-input-focus *display* ,focus ,revert)))))
+  `(progn
+     (grab-keyboard ,xwin)
+     (unwind-protect
+          (progn ,@body)
+       (ungrab-keyboard))))
 
 (defvar *last-unhandled-error* nil
   "If an unrecoverable error occurs, this variable will contain the
@@ -1042,6 +1073,11 @@ sync-all-frame-windows to see the change.")
                                                      (list ".stumpwm.d")))
   "The directory used by stumpwm to store data between sessions.")
 
+(defun data-dir-file (name &optional type)
+  "Return a pathname inside stumpwm's data dir with the specified name and type"
+  (ensure-directories-exist *data-dir*)
+  (make-pathname :name name :type type :defaults *data-dir*))
+
 (defmacro with-data-file ((s file &rest keys &key (if-exists :supersede) &allow-other-keys) &body body)
   "Open a file in StumpWM's data directory. keyword arguments are sent
 directly to OPEN. Note that IF-EXISTS defaults to :supersede, instead
@@ -1051,3 +1087,12 @@ of :error."
      (with-open-file (,s ,(merge-pathnames *data-dir* file)
                          ,@keys)
        ,@body)))
+
+(defmacro move-to-head (list elt)
+   "Move the specified element in in LIST to the head of the list."
+ `(progn
+    (setf ,list (remove ,elt ,list))
+    (push ,elt ,list)))
+
+(define-condition stumpwm-error (error)
+  () (:documentation "Any stumpwm specific error should inherit this."))
