@@ -24,6 +24,7 @@
 (in-package :stumpwm)
 
 (export '(cancel-timer
+	  load-contrib
 	  run-with-timer
 	  stumpwm
 	  timer-p))
@@ -52,15 +53,19 @@ loaded. When CATCH-ERRORS is nil, errors are left to be handled further up. "
 
 (defun error-handler (display error-key &rest key-vals &key asynchronous &allow-other-keys)
   "Handle X errors"
-  ;; ignore asynchronous window errors
-  (if (and asynchronous
-           (find error-key '(xlib:window-error xlib:drawable-error xlib:match-error)))
-      (dformat 4 "Ignoring error: ~s~%" error-key)
-      ;; all other errors are thrown and caught at the top level where
-      ;; stumpwm quits, basically.
-      (if asynchronous
-          (apply 'cerror "Ignore" error-key :display display :error-key error-key key-vals)
-          (apply 'error error-key :display display :error-key error-key key-vals))))
+  (cond 
+    ;; ignore asynchronous window errors
+    ((and asynchronous
+          (find error-key '(xlib:window-error xlib:drawable-error xlib:match-error)))
+     (dformat 4 "Ignoring error: ~s~%" error-key))
+    ((eq error-key 'xlib:access-error)
+     (write-line "Another window manager is running.")
+     (throw :top-level :quit))
+     ;; all other asynchronous errors are printed.
+     (asynchronous
+      (message "Caught Asynchronous X Error: ~s ~s" error-key key-vals))
+     (t
+      (apply 'error error-key :display display :error-key error-key key-vals))))
 
 ;;; Timers
 
@@ -123,26 +128,41 @@ of those expired."
             internal-time-units-per-second)
          0)))
 
+(defun perform-top-level-error-action (c)
+  (ecase *top-level-error-action*
+    (:message
+     (let ((s (format nil "~&Caught '~a' at the top level. Please report this." c)))
+       (write-line s)
+       (print-backtrace)
+       (message "^1*^B~a" s)))
+    (:break (invoke-debugger c))
+    (:abort
+     (throw :top-level (list c (backtrace-string))))))
+
 (defun stumpwm-internal-loop ()
   "The internal loop that waits for events and handles them."
   (loop
      (run-hook *internal-loop-hook*)
      (handler-bind
-         ;;          ((or xlib:window-error xlib:drawable-error) (lambda (c)
-         ;;            ;; Just in case some synchronous window error gets here
-         ;;            ;; (this should be impossible) catch it and ignore it.
-         ;;            (dformat 4 "top level ignore synchronous ~a~%" c))
-         ((error (lambda (c)
-                   (run-hook *top-level-error-hook*)
-                   (ecase *top-level-error-action*
-                     (:message
-                      (let ((s (format nil "~&Caught '~a' at the top level. Please report this." c)))
-                        (write-line s)
-                        (print-backtrace)
-                        (message "^1*^B~a" s)))
-                     (:break (invoke-debugger c))
-                     (:abort
-                      (throw :top-level (list c (backtrace-string))))))))
+         ((xlib:lookup-error (lambda (c)
+                               (if (lookup-error-recoverable-p)
+                                   (recover-from-lookup-error)
+                                   (error c))))
+          (warning #'muffle-warning)
+          ((or serious-condition error)
+           (lambda (c)
+             (run-hook *top-level-error-hook*)
+             (perform-top-level-error-action c)))
+          (t
+           (lambda (c)
+             ;; some other wacko condition was raised so first try
+             ;; what we can to keep going.
+             (cond ((find-restart 'muffle-warning)
+                    (muffle-warning))
+                   ((find-restart 'continue)
+                    (continue)))
+             ;; and if that fails treat it like a top level error.
+             (perform-top-level-error-action c))))
        ;; Note: process-event appears to hang for an unknown
        ;; reason. This is why it is passed a timeout in hopes that
        ;; this will keep it from hanging.
@@ -190,14 +210,10 @@ of those expired."
                ;; we need to do this first because init-screen grabs keys
                (update-modifier-map)
                ;; Initialize all the screens
-               (handler-case
-                   (progn (setf *screen-list* (loop for i in (xlib:display-roots *display*)
-                                                 for n from 0
-                                                 collect (init-screen i n host)))
-                          (xlib:display-finish-output *display*))
-                 (xlib:access-error ()
-                   (write-line "Another window manager is running.")
-                   (throw :top-level :quit)))
+               (setf *screen-list* (loop for i in (xlib:display-roots *display*)
+                                      for n from 0
+                                      collect (init-screen i n host)))
+               (xlib:display-finish-output *display*)
                ;; Load rc file
                (let ((*package* (find-package *default-package*)))
                  (multiple-value-bind (success err rc) (load-rc-file)
