@@ -259,7 +259,11 @@ each directory seperated by a colon."
 
 (defun rehash (&optional (paths (mapcar 'parse-namestring (split-string (getenv "PATH") ":"))))
   "Update the cache of programs in the path stored in @var{*programs-list*} when needed."
-  (let ((dates (mapcar 'portable-file-write-date paths)))
+  (let ((dates (mapcar (lambda (p)
+                         (when (probe-path p)
+                           (portable-file-write-date p)))
+                       paths)))
+    (finish-output)
     (unless (and *path-cache*
                  (equal (path-cache-paths *path-cache*) paths)
                  (equal (path-cache-modification-dates *path-cache*) dates))
@@ -374,26 +378,25 @@ instance. @var{all-groups} overrides this default. Similarily for
            (focus-all win)
            (unless (eq frame old-frame)
              (show-frame-indicator group))))
-       (find-window (group)
-         (find-if (lambda (w)
-                    (apply 'window-matches-properties-p w props))
-                  (group-windows group))))
-    (let*
-        ((screens (if all-screens
-                      *screen-list*
-                      (list (current-screen))))
-         (win
-          ;; If no qualifiers are set don't bother looking for a match.
-          ;; search all groups
-          (if all-groups
-              (loop named outer
-                    for s in screens
-                    do (loop
-                        for g in (screen-groups s)
-                        for win = (find-window g)
-                        when win
-                        do (return-from outer win)))
-              (find-window (current-group)))))
+       (sort-windows-by-group (winlist)
+         (stable-sort (sort winlist #'< :key #'window-number)
+                      #'< :key #'(lambda (w) (group-number (window-group w)))))
+       (find-window (winlist)
+         (let* ((match (remove-if-not #'(lambda (w)
+                                          (apply 'window-matches-properties-p w props))
+                                      winlist))
+                (match-sorted (sort-windows-by-group match))
+                (rest (member (current-window) match-sorted)))
+           (if (<= (length rest) 1) ; current win not a match or no matches left
+               (car match-sorted)
+               (cadr rest)))))
+    (let* ((screens (if all-screens
+                        *screen-list*
+                        (list (current-screen))))
+           (winlist (if all-groups
+                        (mapcan (lambda (s) (screen-windows s)) screens)
+                        (group-windows (current-group))))
+           (win (find-window winlist)))
       (if win
           (goto-win win)
           (run-shell-command cmd)))))
@@ -431,3 +434,15 @@ submitting the bug report."
      (defcommand ,name () ()
        (run-or-raise ,command ,props))
      (define-key ,map ,key ,(string-downcase (string name)))))
+
+(defcommand show-window-properties () ()
+  "Shows the properties of the current window. These properties can be
+used for matching windows with run-or-raise or window placement
+rules."
+  (let ((w (current-window)))
+    (message-no-timeout "class: ~A~%instance: ~A~%type: :~A~%role: ~A~%title: ~A"
+                        (window-class w)
+                        (window-res w)
+                        (string (window-type w))
+                        (window-role w)
+                        (window-title w))))
