@@ -25,12 +25,16 @@
 
 (in-package #:stumpwm)
 
-(export '(def-window-attr
+(export '(*default-window-name*
+          def-window-attr
           set-normal-gravity
           set-maxsize-gravity
           set-transient-gravity
           set-window-geometry))
 
+(defvar *default-window-name* "Unnamed"
+  "The name given to a window that does not supply its own name.")
+
 ;; Urgency / demands attention
 
 (defun register-urgent-window (window)
@@ -52,9 +56,15 @@
 (defun window-clear-urgency (window)
   "Clear the urgency bit and/or _NET_WM_STATE_DEMANDS_ATTENTION on
 WINDOW"
-  (and (xlib:wm-hints (window-xwin window))
-       (let ((flags (xlib:wm-hints-flags (xlib:wm-hints (window-xwin window)))))
-         (setf flags (logand (lognot 256) flags))))
+  (let* ((hints (xlib:wm-hints (window-xwin window)))
+         (flags (when hints (xlib:wm-hints-flags hints))))
+    (when flags
+      (setf (xlib:wm-hints-flags hints)
+            ;; XXX: as of clisp 2.46 flags is a list, not a number.
+            (if (listp flags)
+                (remove :urgency flags)
+                (logand (lognot 256) flags)))
+      (setf (xlib:wm-hints (window-xwin window)) hints)))
   (remove-wm-state (window-xwin window) :_NET_WM_STATE_DEMANDS_ATTENTION)
   (unregister-urgent-window window))
 
@@ -63,13 +73,21 @@ WINDOW"
 _NET_WM_STATE_DEMANDS_ATTENTION set"
   (let* ((hints (xlib:wm-hints (window-xwin window)))
          (flags (when hints (xlib:wm-hints-flags hints))))
-    (or (and flags (logtest 256 flags))
+    ;; XXX: as of clisp 2.46 flags is a list, not a number.
+    (or (and flags (if (listp flags)
+                       (find :urgency flags)
+                       (logtest 256 flags)))
         (find-wm-state (window-xwin window) :_NET_WM_STATE_DEMANDS_ATTENTION))))
 
 (defun only-urgent (windows)
   "Return a list of all urgent windows on SCREEN"
   (remove-if-not 'window-urgent-p (copy-list windows)))
 
+(defcommand next-urgent () ()
+            "Jump to the next urgent window"
+            (and (screen-urgent-windows (current-screen))
+                 (focus-all (first (screen-urgent-windows (current-screen))))))
+
 ;; Since StumpWM already uses the term 'group' to refer to Virtual Desktops,
 ;; we'll call the grouped windows of an application a 'gang'
 
@@ -182,7 +200,8 @@ _NET_WM_STATE_DEMANDS_ATTENTION set"
       (case *window-name-source*
         (:resource-name (window-res window))
         (:class (window-class window))
-        (t (window-title window)))))
+        (t (window-title window)))
+      *default-window-name*))
 
 (defun window-id (window)
   (xlib:window-id (window-xwin window)))
@@ -247,6 +266,26 @@ _NET_WM_STATE_DEMANDS_ATTENTION set"
 ;;                           0 0 300 200 t)
 ;;      (xlib:clear-area (window-parent window)))))
 
+(defun get-normalized-normal-hints (xwin)
+  (macrolet ((validate-hint (fn)
+               (setf fn (intern (concatenate 'string (string :wm-size-hints-) (string fn)) :xlib))
+               `(setf (,fn hints) (and (,fn hints)
+                                       (plusp (,fn hints))
+                                       (,fn hints)))))
+    (let ((hints (xlib:wm-normal-hints xwin)))
+      (when hints
+        (validate-hint :min-width)
+        (validate-hint :min-height)
+        (validate-hint :max-width)
+        (validate-hint :max-height)
+        (validate-hint :base-width)
+        (validate-hint :base-height)
+        (validate-hint :width-inc)
+        (validate-hint :height-inc)
+        (validate-hint :min-aspect)
+        (validate-hint :max-aspect))
+      hints)))
+
 (defun xwin-net-wm-name (win)
   "Return the netwm wm name"
   (let ((name (xlib:get-property win :_NET_WM_NAME)))
@@ -408,7 +447,7 @@ _NET_WM_STATE_DEMANDS_ATTENTION set"
 
 (defun xwin-maxsize-p (win)
   "Returns T if WIN specifies maximum dimensions."
-  (let ((hints (xlib:wm-normal-hints win)))
+  (let ((hints (get-normalized-normal-hints win)))
     (and hints (or (xlib:wm-size-hints-max-width hints)
                    (xlib:wm-size-hints-max-height hints)
                    (xlib:wm-size-hints-min-aspect hints)
@@ -559,11 +598,11 @@ than the root window's width and height."
       (t
        ;; if they have inc hints then start with the size and adjust
        ;; based on those increments until the window fits in the frame
-       (when hints-inc-x
+       (when (and hints-inc-x (plusp hints-inc-x))
          (let ((w (or hints-width (window-width win))))
            (setf width (+ w (* hints-inc-x
                                (+ (floor (- fwidth w) hints-inc-x)))))))
-       (when hints-inc-y
+       (when (and hints-inc-y (plusp hints-inc-y))
          (let ((h (or hints-height (window-height win))))
            (setf height (+ h (* hints-inc-y
                                 (+ (floor (- fheight h -1) hints-inc-y)))))))))
@@ -618,7 +657,13 @@ than the root window's width and height."
             (setf (xlib:drawable-width (window-parent win)) (- (frame-width frame)
                                                                (* 2 (xlib:drawable-border-width (window-parent win))))
                   (xlib:drawable-height (window-parent win)) (- (frame-display-height (window-group win) frame)
-                                                                (* 2 (xlib:drawable-border-width (window-parent win))))))))))
+                                                                (* 2 (xlib:drawable-border-width (window-parent win)))))))
+      ;; update the "extents"
+      (xlib:change-property (window-xwin win) :_NET_FRAME_EXTENTS
+                            (list wx wy
+                                  (- (xlib:drawable-width (window-parent win)) width wx)
+                                  (- (xlib:drawable-height (window-parent win)) height wy))
+                            :cardinal 32))))
 
 (defun find-free-window-number (group)
   "Return a free window number for GROUP."
@@ -683,7 +728,7 @@ than the root window's width and height."
 
 (defun xwin-grab-keys (win)
   (labels ((grabit (w key)
-             (let* ((code (xlib:keysym->keycodes *display* (key-keysym key))))
+             (loop for code in (multiple-value-list (xlib:keysym->keycodes *display* (key-keysym key))) do
                ;; some keysyms aren't mapped to keycodes so just ignore them.
                (when code
                  ;; Some keysyms, such as upper case letters, need the
@@ -757,7 +802,7 @@ than the root window's width and height."
    :res (xwin-res-name xwin)
    :role (xwin-role xwin)
    :type (xwin-type xwin)
-   :normal-hints (xlib:wm-normal-hints xwin)
+   :normal-hints (get-normalized-normal-hints xwin)
    :state +iconic-state+
    :plist (make-hash-table)
    :unmap-ignores 0))
@@ -877,7 +922,7 @@ than the root window's width and height."
       (setq preferred-frame
             (handler-case
                 (funcall *new-window-preferred-frame* window)
-              (error (c)
+              (t (c)
                 (message "^1*^BError while calling ^b^3**new-window-preferred-frame*^1*^B: ^n~a" c)
                 default))))
     (cond
@@ -976,7 +1021,7 @@ needed."
             (window-res window) (xwin-res-name (window-xwin window))
             (window-role window) (xwin-role (window-xwin window))
             (window-type window) (xwin-type (window-xwin window))
-            (window-normal-hints window) (xlib:wm-normal-hints (window-xwin window))
+            (window-normal-hints window) (get-normalized-normal-hints (window-xwin window))
             (window-number window) (find-free-window-number (window-group window))
             (window-state window) +iconic-state+
             (xwin-state (window-xwin window)) +iconic-state+
@@ -1050,6 +1095,7 @@ needed."
           (delete window (screen-urgent-windows screen)))
     (dformat 1 "destroy window ~a~%" screen)
     (dformat 3 "destroying parent window~%")
+    (dformat 7 "parent window is ~a~%" (window-parent window))
     (xlib:destroy-window (window-parent window))))
 
 (defun move-window-to-head (group window)
@@ -1057,8 +1103,7 @@ needed."
   (declare (type group group))
   (declare (type window window))
                                         ;(assert (member window (screen-mapped-windows screen)))
-  (setf (group-windows group) (delete window (group-windows group)))
-  (push window (group-windows group))
+  (move-to-head (group-windows group) window)
   (netwm-update-client-list-stacking (group-screen group)))
 
 (defun no-focus (group last-win)
@@ -1238,7 +1283,7 @@ with broken (non-NETWM) fullscreen implemenations, such as any program
 using SDL."
   (update-fullscreen (current-window) 2))
 
-(defcommand pull-window-by-number (n &optional (group (current-group))) 
+(defcommand pull-window-by-number (n &optional (group (current-group)))
                                   ((:window-number "Pull: "))
   "Pull window N from another frame into the current frame and focus it."
   (let ((win (find n (group-windows group) :key 'window-number :test '=)))
@@ -1265,6 +1310,20 @@ is using the number, then the windows swap numbers. Defaults to current group."
 
 (defcommand-alias number renumber)
 
+(defcommand repack-window-numbers (&optional preserved) ()
+  "Ensure that used window numbers do not have gaps; ignore PRESERVED window numbers."
+  (let* ((group (current-group))
+	 (windows (sort-windows group)))
+    (loop for w in windows
+	  do (unless (find (window-number w) preserved)
+	       (setf
+		 (window-number w)
+		 (find-free-number
+		   (remove
+		     (window-number w)
+		     (mapcar 'window-number windows))
+		   0))))))
+
 (defcommand gravity (gravity) ((:gravity "Gravity: "))
   (when (current-window)
     (setf (window-gravity (current-window)) gravity)
@@ -1428,7 +1487,7 @@ be used to override the default window formatting."
                 :role (and (not (equal role "")) role))
           *window-placement-rules*)))
 
-(defcommand remember (lock title) 
+(defcommand remember (lock title)
                      ((:y-or-n "Lock to group? ")
                       (:y-or-n "Use title? "))
   "Make a generic placement rule for the current window. Might be too specific/not specific enough!"
@@ -1455,3 +1514,9 @@ be used to override the default window formatting."
   (setf *window-placement-rules* (read-dump-from-file file)))
 
 (defcommand-alias restore-rules restore-window-placement-rules)
+
+(defcommand info (&optional (fmt *window-info-format*)) ()
+  "Display information about the current window."
+  (if (current-window)
+      (message "~a" (format-expand *window-formatters* fmt (current-window)))
+      (message "No Current Window")))
